;(function ( w, doc, undefined ) {
	'use strict';

	/**
	 * Local object for method references,
	 * define script metadata, and other
	 * global variables.
	 */
	var ARIAmodal = {};
	w.ARIAmodal   = ARIAmodal;

	ARIAmodal.NS      = 'ARIAmodal';
	ARIAmodal.AUTHOR  = 'Scott O\'Hara';
	ARIAmodal.VERSION = '3.4.0';
	ARIAmodal.LICENSE = 'https://github.com/scottaohara/accessible_modal_window/blob/master/LICENSE';

	var activeClass = 'modal-open';
	var body = doc.body;
	var main = doc.getElementsByTagName('main')[0] || body;

	var modal = doc.querySelectorAll('[data-modal]');
	var children = [];

	var initialTrigger;
	var activeModal;
	var useAriaModal = false;
	var returnToBody = false;

	var firstClass = 'js-first-focus';
	var lastClass = 'js-last-focus';

	var closeFocus;
	var modalButtons;
	var modalOptions = {};

	var tabFocusElements = 'button:not([hidden]):not([disabled]), [href]:not([hidden]), input:not([hidden]):not([type="hidden"]):not([disabled]), select:not([hidden]):not([disabled]), textarea:not([hidden]):not([disabled]), [tabindex="0"]:not([hidden]):not([disabled]), summary:not([hidden]), [contenteditable]:not([hidden]), audio[controls]:not([hidden]), video[controls]:not([hidden])';

	ARIAmodal.setFocus = function(arg) {
		if (typeof arg === 'string') {
			arg = document.getElementById(arg);
		}
		setTimeout(function () {
			try {
				arg.focus();
			}
			catch (err) {
				console.log('ARIAmodal.setFocus: ' + err.message);
			}
		}, 100);
	};

	ARIAmodal.getChildren = function () {
		var childNodes = doc.querySelectorAll('body > *:not([data-modal])');
		for (var i = 0; i < childNodes.length; i++) {
			if (childNodes[i].tagName.toLowerCase() !== 'script' && childNodes[i].tagName.toLowerCase() !== 'br') {
				children.push(childNodes[i]);
			}
		}
	};


	/**
	 * Function to place the modal dialog(s) as the first child(ren)
	 * of the body element, so tabbing backwards will move focus
	 * into the browser's chrome.
	 */
	ARIAmodal.organizeDOM = function () {
		var refEl = body.firstElementChild || null;
		var i;

		for ( i = 0; i < modal.length; i++ ) {
			body.insertBefore( modal[i], refEl );
		}
	};


	/**
	 * Setup the necessary attributes and child elements for the
	 * modal dialogs.
	 */
	ARIAmodal.setupModal = function () {
		var self;
		var i;

		for ( i = 0; i < modal.length; i++ ) {
			var self = modal[i];
			var modalType   = self.getAttribute('data-modal');
			var getClass    = self.getAttribute('data-modal-class') || 'a11y-modal';
			var heading     = self.querySelector('h1, h2, h3, h4, h5, h6');
			var modalLabel  = self.getAttribute('data-modal-label');
			var hideHeading = self.hasAttribute('data-modal-hide-heading');
			var modalDesc   = self.querySelector('[data-modal-description]');
			var modalDoc    = self.querySelector('[data-modal-document]');

			/**
			 * Check to see if this is meant to be an alert or normal dialog.
			 * Supply the appropriate role.
			 */
			if ( modalType === 'alert' ) {
				self.setAttribute('role', 'alertdialog');
			}
			else {
				self.setAttribute('role', 'dialog');
			}

			/**
			 * Set either the default dialog class or a class passed
			 * in from the data-modal-class attribute.
			 */
			self.classList.add(getClass);

			/**
			 * Modal dialogs need to be hidden by default.
			 *
			 * To ensure they stay hidden, even if CSS is disabled, or purposefully
			 * turned off, apply a [hidden] attribute to the dialogs.
			 */
			self.hidden = true;

			/**
			 * When a modal dialog is opened, the dialog itself
			 * should be focused. Set a tabindex="-1" to allow this
			 * while keeping the container out of the focus order.
			 */
			self.tabIndex = '-1';


			/**
			 * Older versions of NVDA used to automatically turn on forms mode
			 * when a user entered a modal dialog. A role="document", surrounding
			 * the content of the dialog would allow non-form dialogs to be
			 * navigated correctly by the virtual cursor.
			 *
			 * If a dialog needs to be compatible with older NVDA, look for
			 * a data-modal-document, and give that a role=document.
			 */
			if ( modalDoc ) {
				modalDoc.setAttribute('role', 'document');
			}


			/**
			 * Modal dialogs need at least one actionable item
			 * to close them...
			 */
			//ARIAmodal.setupModalCloseBtn(self, getClass, modalType);

			/**
			 * The aria-modal attribute currently makes it difficult to navigate
			 * through the contents of a modal dialog with VoiceOver.
			 *
			 * Up/down arrows do not have access to all content, and
			 * using VO + left/right also do not have access to all
			 * content, but do have access to different content then
			 * up/down arrows alone.
			 *
			 * Note: The VoiceOver issues should be fixed with the release
			 * of Safari 12.
			 *
			 * Additionally, NVDA will mostly respect the aria-modal attribute
			 * with one notable bug, where if a user navigates to the address
			 * bar via NVDA key + F6, a user can re-enter document that is obscured
			 * by the open dialog, and can navigate the content 'beneath' the
			 * dialog with arrow keys, or quick keys.
			 *
			 * Using the tab key inconsistently returns a user to the modal
			 * dialog's contents, or may produce no keyboard focus change.
			 *
			 * This attribute can be added to a particular dialog if the
			 * dialog has a data-aria-modal attribute set.
			 */
			if ( self.hasAttribute('data-aria-modal') ) {
				self.setAttribute('aria-modal', 'true');
			}


			/**
			 * Do a check to see if there is an element flagged to be the
			 * description of the modal dialog.
			 */
			if ( modalDesc ) {
				modalDesc.id = modalDesc.id || 'md_desc_' + Math.floor(Math.random() * 999) + 1;
				self.setAttribute('aria-describedby', modalDesc.id);
			}

			/**
			 * Check for a heading to set the accessible name of the dialog,
			 * or if an aria-label should be set to the dialog instead.
			 */
			if ( modalLabel ) {
				self.setAttribute('aria-label', modalLabel);
			}
			else {
				if ( heading ) {
					var makeHeading = self.id + '_heading';
					heading.classList.add(getClass + '__heading');
					heading.id = makeHeading;

					/**
					 * Set an aria-labelledby to the modal dialog container.
					 */
					self.setAttribute('aria-labelledby', makeHeading);

					if ( heading.hasAttribute('data-autofocus') ) {
						heading.tabIndex = '-1';
					}

					self.setAttribute('aria-describedby', 'aria_modal_text');
				}
				else {
					console.warn('Dialogs should have their purpose conveyed by a heading element (h1).');
				}
			}

			/**
			 * If a dialog has a data-modal-hide-heading attribute, then that means this
			 * dialog's heading should be visually hidden.
			 */
			if ( hideHeading ) {
				self.querySelector('#' + heading.id).classList.add('at-only');
			}

			/**
			 * Get all focusable elements from within a dialog and set the
			 * first and last elements to have respective classes for later looping.
			 */
			/*
			var focusable = self.querySelectorAll(tabFocusElements);
			focusable[0].classList.add(firstClass);
			focusable[focusable.length - 1].classList.add(lastClass);
			*/

			self.addEventListener('click', ARIAmodal.handleButtonClick, false);
		} // end for loop
	}; // ARIAmodal.setupModal


	ARIAmodal.handleButtonClick = function (event) {
		var target = event.target;
		if (target.tagName.toLowerCase() !== 'button') {
			return;
		}
		event.preventDefault();
		event.stopPropagation();
		// Set this to false so that closeModal() doesn't try to set the focus.
		// Setting focus should be handled by the onClick handler for the button.
		closeFocus = false;
		ARIAmodal.closeModal();
		var buttonIndex = Number(event.target.dataset.buttonindex);
		setTimeout(function () {
			modalButtons[buttonIndex].onClick();
		}, 100);
	};
 
	/**
	 * Actions
	 */
	ARIAmodal.openModal = function (/*e,*/ /*autoOpen*/ options) {

		if (typeof options !== 'object' || typeof options.modalId !== 'string') {
			console.log('ARIAmodal.openModal: bad options argument: ', options);
			return;
		}

		modalOptions = options;

		if (typeof options.closeFocus === 'string') {
			closeFocus = options.closeFocus;
		}
		else {
			closeFocus = 'body';
		}

		if (typeof options.text === 'string') {
			if (options.text !== '') {
				document.getElementById('aria_modal_text').innerText = options.text;
			}
			else {
				document.getElementById('aria_modal_text').innerText = '';
			}
		}
		else {
			document.getElementById('aria_modal_text').innerText = '';
		}

		document.getElementById('aria_modal_heading').innerText = options.title;

		//var showCloseBtn = (options.closeBtn === false) ? false : true;
		//var closeBtnEl = document.getElementById('modal-close-button');

		
		/*
		if (!showCloseBtn) {
			closeBtnEl.setAttribute('aria-hidden', 'true');
			closeBtnEl.setAttribute('tabindex', '-1');
		}
		else {
			closeBtnEl.setAttribute('aria-hidden', 'false');
			closeBtnEl.setAttribute('tabindex', '0');
		}
		*/
		

		var i;
		//var getTargetModal = options.modalId;
		// Update the activeModal
		activeModal = doc.getElementById(options.modalId);

		if (activeModal === null) {
			return;
		}

		//var focusTarget = activeModal; // default to the modal dialog container
		//var getAutofocus = activeModal.querySelector('[autofocus]') || activeModal.querySelector('[data-autofocus]');

		useAriaModal = activeModal.hasAttribute('aria-modal');

		/**
		 * If a modal dialog was auto-opened, then a user should
		 * be returned to the top of the document when the modal
		 * is closed, so that they do not have to figure out where
		 * they've been placed in the DOM
		 */
		/*
		if ( autoOpen ) {
			returnToBody = true;
			e.originalEvent.preventDefault();
		}
		*/

		/**
		 * If a modal was auto-opened on page load, then the
		 * following do not apply.
		 */
		//if ( !autoOpen ) {
			/**
			 * In case these are links, negate default behavior and just
			 * do what this script tells these triggers to do.
			 */
			//e.preventDefault();

			/**
			 * Keep track of the trigger that opened the initial dialog.
			 */
			//initialTrigger = this.id;
		//}

		/**
		 * If a modal dialog contains an that is meant to be autofocused,
		 * then focus should be placed on that element (likely form control),
		 * instead of the wrapping dialog container.
		 *
		 * If a dialog has an attribute indicating the close button should
		 * be autofocused, focus the first close button found.
		 */
		/*
		if ( getAutofocus ) {
			focusTarget = getAutofocus;
		}
		else if ( activeModal.hasAttribute('data-modal-close-focus') ) {
			focusTarget = activeModal.querySelector('[class*="close-btn"]');
		}
		*/

		// Add buttons
		if (Array.isArray(options.buttons)) {
			var modalButtonsDiv = document.getElementById('aria_modal_buttons');
			modalButtons = options.buttons;
			for (var i = 0; i < modalButtons.length; i++) {
				var newBtn = document.createElement('button');
				newBtn.id = 'modal-button-' + i;
				newBtn.innerText = modalButtons[i].text;
				newBtn.dataset.buttonindex = i;
				modalButtonsDiv.appendChild(newBtn);
			}
		}
		else {
			modalButtons = undefined;
		}

		var focusable = activeModal.querySelectorAll(tabFocusElements);
		focusable[0].classList.add(firstClass);
		focusable[focusable.length - 1].classList.add(lastClass);

		/**
		 * Do a check to see if a modal is already open.
		 * If not, then add a class to the body as a check
		 * for other functions and set contents other than
		 * the opened dialog to be hidden from screen readers
		 * and to not accept tab focus, nor for their child elements.
		 */
		if ( !body.classList.contains(activeClass) ) {
			body.classList.add(activeClass);

			for (i = 0; i < children.length; i++) {

				//if (children[i].getAttribute('inert')) {
				if (children[i].inert === true) {
					children[i].setAttribute('data-keep-inert', '');
				}
				else {
					//children[i].setAttribute('inert', 'true');
					children[i].inert = true;
				}

				// brb - Even though we should be adding aria-modal="true" to
				// all modal dialogs and thus this code should never run I am
				// commenting it out anyway just to make sure.
				/*
				if ( !useAriaModal ) {
					if ( children[i].hasAttribute('aria-hidden') ) {
						children[i].setAttribute('data-keep-hidden', children[i].getAttribute('aria-hidden') );
					}

					// brb: We don't need to do this because the the 'inert'
					// polyfil does this for us.
					
					children[i].setAttribute('aria-hidden', 'true');
					
				}
				*/
				
			}
		}
		else {
			console.warn('It is not advised to open dialogs from within other dialogs. Instead consider replacing the contents of this dialog with new content. Or providing a stepped, or tabbed interface within this dialog.');
		}

		activeModal.removeAttribute('hidden');

		// Mostly reliable fix for iOS issue where VO focus is not moved
		// to the dialog on open. Credit to Thomas Jaggi - codepen.io/backflip
		// for the fix.
		requestAnimationFrame(function () {
			ARIAmodal.setFocus(focusable[0]);
	    	//focusTarget.focus();
	  	});

		doc.addEventListener('click', ARIAmodal.outsideClose, false);
		doc.addEventListener('touchend', ARIAmodal.outsideClose, false);
		doc.addEventListener('keydown', ARIAmodal.keyEvents, false);

		if (typeof options.top === 'number') {
			activeModal.style.top = options.top + 'px';
		}

		return [initialTrigger, activeModal, returnToBody];
	};


	/**
	 * Function for closing a modal dialog.
	 * Remove inert, and aria-hidden from non-dialog parent elements.
	 * Remove activeClass from body element.
	 * Focus the appropriate element.
	 */
	ARIAmodal.closeModal = function (e) {
	
		var trigger = doc.getElementById(initialTrigger) || null;
		var i;
		var m;

		doc.removeEventListener('click', ARIAmodal.outsideClose, false);
		doc.removeEventListener('touchend', ARIAmodal.outsideClose, false);
		doc.removeEventListener('keydown', ARIAmodal.keyEvents, false);

		// remove any custom buttons
		var modalButtonsEl = document.getElementById('aria_modal_buttons');
		var modalButtonsChildren = modalButtonsEl.children;

		while (modalButtonsChildren.length > 0) {
			modalButtonsEl.removeChild(modalButtonsChildren[0]);
		}

		/**
		 * Loop through all the elements that were hidden to
		 * screen readers, and had inert to negate their
		 * children from being focusable.
		 */
		for (i = 0; i < children.length; i++) {
			
			if (!children[i].hasAttribute('data-keep-inert')) {
				//children[i].removeAttribute('inert');
				children[i].inert = false;
			}
			
			children[i].removeAttribute('data-keep-inert');

			if ( children[i].getAttribute('data-keep-hidden') ) {
				children[i].setAttribute('aria-hidden', children[i].getAttribute('data-keep-hidden') );
			}
			else {
				children[i].removeAttribute('aria-hidden');
			}
		}

		/**
		 * When a modal closes:
		 * the modal-open flag on the body can be removed,
		 * and the modal should be reset to hidden.
		 */
		body.classList.remove(activeClass);

		for ( m = 0; m < modal.length; m++ ) {
			if ( !modal[m].hasAttribute('hidden') ) {
				modal[m].hidden = true;
			}
		}

		/**
		 * Return focus to the trigger that opened the modal dialog.
		 * If the trigger doesn't exist for some reason, move focus to
		 * either the <main>, or <body> instead.
		 * Reset initialTrigger and activeModal since everything should be reset.
		 */
		
		if (closeFocus === false) {
			// we don't set focus if closeFocus explicitly set to 'false'
		}
		else if (typeof closeFocus === 'string') {
			setTimeout(function () {
				ARIAmodal.setFocus(closeFocus);
				//document.getElementById(closeFocus).focus();
				closeFocus = undefined;
			}, 100);
		}
		else if ( trigger !== null ) {
			ARIAmodal.setFocus(trigger);
		}
		else {
			if ( main && !returnToBody ) {
				main.tabIndex = -1;
				ARIAmodal.setFocus(main);
			}
			else {
				body.tabIndex = -1;
				ARIAmodal.setFocus(body);
			}
		}

		initialTrigger = undefined;
		activeModal = undefined;
		returnToBody = false;

		return [initialTrigger, activeModal, returnToBody];
	};


	/**
	 * Keyboard controls for when the modal dialog is open.
	 * ESC should close the dialog (when not an alert)
	 */
	ARIAmodal.keyEvents = function ( e ) {
		var keyCode  = e.keyCode || e.which;
		var escKey   = 27;
		//var enterKey = 13;
		//var spaceKey = 32;
		var tabKey   = 9;

		if ( body.classList.contains(activeClass) ) {
			switch ( keyCode ) {
				case escKey:
					e.preventDefault();
					ARIAmodal.closeModal();
					if (typeof modalOptions.onEsc === 'function') {
						modalOptions.onEsc();
					}
					break;

				default:
					break;
			}

			if ( body.classList.contains(activeClass) ) {
				// Get first and last focusable elements from activeModal
				var firstFocus = activeModal.querySelector('.' + firstClass);
				var lastFocus = activeModal.querySelector('.' + lastClass);
			}

			if ( doc.activeElement.classList.contains(lastClass) ) {
				if ( keyCode === tabKey && !e.shiftKey ) {
					e.preventDefault();
					ARIAmodal.setFocus(firstFocus);
				}
			}

			if ( doc.activeElement.classList.contains(firstClass) ) {
				if ( keyCode === tabKey && e.shiftKey ) {
					e.preventDefault();
					ARIAmodal.setFocus(lastFocus);
				}
			}
		}
	}; // ARIAmodal.keyEvents()


	/**
	 * If a dialog is opened and a user mouse clicks or touch screen taps outside
	 * the visible bounds of the dialog content (onto the overlay 'screen') then
	 * the dialog should run the close function.
	 *
	 * NOTE: This will not work if specified as an alert dialog!
	 *  
	 */
	ARIAmodal.outsideClose = function (e) {
		if ( body.classList.contains(activeClass) && !e.target.hasAttribute('data-modal-open') ) {
			var isClickInside = activeModal.contains(e.target);

			if (!isClickInside && activeModal.getAttribute('role') !== 'alertdialog') {
				e.preventDefault();
				e.stopPropagation();
				ARIAmodal.closeModal();
			}
		}
	}; // ARIAmodal.outsideClose()


	/**
	 * Initialize modal functions.
	 * If expanding this script, put
	 * additional initialize functions here.
	 */
	ARIAmodal.init = function () {
		ARIAmodal.getChildren();
		ARIAmodal.organizeDOM();
		ARIAmodal.setupModal();
	};


	/**
	 * Go go JavaScript!
	 */
	ARIAmodal.init();

})( window, document );
